/* debug.c                            written by Kenji Kise, 2001
 * A part of SimpleScalar Simulator.          Developing on Linux
 ****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "define.h"

extern struct evaluate e;

unsigned int get_hex(char *buf){
  int i;
  unsigned int temp=0;
  unsigned int sum=0;
  for(i=0; i<16; i++){
    switch(*(buf+i)){
    case '0': temp=0; break;
    case '1': temp=1; break;
    case '2': temp=2; break;
    case '3': temp=3; break;
    case '4': temp=4; break;
    case '5': temp=5; break;
    case '6': temp=6; break;
    case '7': temp=7; break;
    case '8': temp=8; break;
    case '9': temp=9; break;
    case 'a': temp=10;break;
    case 'b': temp=11;break;
    case 'c': temp=12;break;
    case 'd': temp=13;break;
    case 'e': temp=14;break;
    case 'f': temp=15;break;
    default : return sum;
    }
    sum = sum * 16 + temp;
  }
  return sum;
}

void print_status(struct architecture_state *as){
  char buf[256];

  getasm(as, buf);
  printf("TC:%qd  ", e.tc);
  printf("PC:0x%08x: IR:0x%02x-%08x %s\n",
      as->pc, as->ir.op, as->ir.low, buf);
  printf("fcc:%08x  hi:%08x  lo:%08x\n", 
	 as->fcc, as->hi, as->lo);

  printf("$zero r0 %08x %08x: ", as->r[0], as->f[0] );
  printf("$at   r1 %08x %08x\n", as->r[1], as->f[1] );
  printf("$v0   r2 %08x %08x: ", as->r[2], as->f[2] );
  printf("$v1   r3 %08x %08x\n", as->r[3], as->f[3] );
  printf("$a0   r4 %08x %08x: ", as->r[4], as->f[4] );
  printf("$a1   r5 %08x %08x\n", as->r[5], as->f[5] );
  printf("$a2   r6 %08x %08x: ", as->r[6], as->f[6] );
  printf("$a3   r7 %08x %08x\n", as->r[7], as->f[7] );
  printf("$t0   r8 %08x %08x: ", as->r[8], as->f[8] );
  printf("$t1   r9 %08x %08x\n", as->r[9], as->f[9] );
  printf("$t2  r10 %08x %08x: ", as->r[10],as->f[10]);
  printf("$t3  r11 %08x %08x\n", as->r[11],as->f[11]);
  printf("$t4  r12 %08x %08x: ", as->r[12],as->f[12]);
  printf("$t5  r13 %08x %08x\n", as->r[13],as->f[13]);
  printf("$t6  r14 %08x %08x: ", as->r[14],as->f[14]);
  printf("$t7  r15 %08x %08x\n", as->r[15],as->f[15]);
  printf("$s0  r16 %08x %08x: ", as->r[16],as->f[16]);
  printf("$s1  r17 %08x %08x\n", as->r[17],as->f[17]);
  printf("$s2  r18 %08x %08x: ", as->r[18],as->f[18]);
  printf("$s3  r19 %08x %08x\n", as->r[19],as->f[19]);
  printf("$s4  r20 %08x %08x: ", as->r[20],as->f[20]);
  printf("$s5  r21 %08x %08x\n", as->r[21],as->f[21]);
  printf("$s6  r22 %08x %08x: ", as->r[22],as->f[22]);
  printf("$s7  r23 %08x %08x\n", as->r[23],as->f[23]);
  printf("$t8  r24 %08x %08x: ", as->r[24],as->f[24]);
  printf("$t9  r25 %08x %08x\n", as->r[25],as->f[25]);
  printf("$k0  r26 %08x %08x: ", as->r[26],as->f[26]);
  printf("$k1  r27 %08x %08x\n", as->r[27],as->f[27]);
  printf("$gp  r28 %08x %08x: ", as->r[28],as->f[28]);
  printf("$sp  r29 %08x %08x\n", as->r[29],as->f[29]);
  printf("$s8  r30 %08x %08x: ", as->r[30],as->f[30]);
  printf("$ra  r31 %08x %08x\n", as->r[31],as->f[31]);
  fflush(stdout);
}
 
void debugmode(struct architecture_state *as){
  static int stop=1;
  int i, j;
  static ULL stop_count=0;
  char *buf =(char *)calloc(256,1);
  
  if(e.tc==stop_count) stop=1;
  if(stop==0) return;
  if(stop==1){
    print_status(as);
    printf("(db)");

    fgets(buf, 256, stdin);
    while(*buf != (char)NULL){
      if(*(buf)=='\n') break;

      if(!strncmp(buf,"q",1)){
	printf("Exit\n"); fflush(stdout); exit(0);
      }
      
      if(!strncmp(buf,"h",1)){ /* Print help message */
	printf("  command\n");
	printf("  c followd by number:  stop specified tc\n");
	printf("  x followd by address: examine specified address\n");
	printf("  q: quit the simulator\n");
      }

      if(*(buf)=='c'){ /* stop by total count,TC */
	buf++;
	stop_count = atol(buf);
	printf("stop_tc = %qd\n", stop_count);
	stop=0;
	break;
      }

      if(*(buf)=='x'){
	ADDR_TYPE tmp;
	ADDR_TYPE address;
	address = get_hex(buf+1);
	for(j=0; j<20; j++){
	  printf("0x%08x: ",address);
	  for(i=0; i<4; i++){
	    load_nbyte(4, SIGNED, address, (int *)&tmp);
	    printf("0x%08x ", tmp);
	    address +=4;
	  }
	  printf("\n");
	}
      }
      printf("\n"); printf("(db)");
      fgets(buf,256,stdin);
    }
  }
}
