/***** 1. type definitions *****/
typedef unsigned long long      ULL;
typedef unsigned int      ADDR_TYPE;
typedef          int      DATA_TYPE;
typedef struct{int op; int low;} INST_TYPE;

/***** 2. Constant Definition *****/
#define PAGE_SIZE          4096       /*  4KB page  */
#define BLOCK_SIZE         0x00010000 /* 64KB block */
#define BLOCK_TABLE_SIZE   0x00008000 
#define BLOCK_MASK         0x0000FFFF
#define STACK_BASE         0x7fffc000
#define MAX_ENVIRON        16384
#define SS_TEXT_BASE       0x00400000

#define SIGNED 1
#define UNSIGN 2

#define __J	0x01
#define __JAL 	0x02 
#define __JR 	0x03 
#define __JALR 	0x04 
#define __BEQ 	0x05 
#define __BNE 	0x06 
#define __BLEZ 	0x07 
#define __BGTZ 	0x08 
#define __BLTZ 	0x09 
#define __BGEZ 	0x0a 
#define __BC1F 	0x0b
#define __BC1T 	0x0c

#define __LB 	0x20
#define __LBU 	0x22 
#define __LH 	0x24 
#define __LHU 	0x26 
#define __LW 	0x28 
#define __DLW 	0x29 
#define __L_S 	0x2a 
#define __L_D 	0x2b 
#define __LWL 	0x2c 
#define __LWR 	0x2d 

#define __SB 	0x30 
#define __SH 	0x32 
#define __SW 	0x34 
#define __S_S 	0x36 
#define __DSW 	0x35 
#define __S_D 	0x37 
#define __DSZ 	0x38 
#define __SWL 	0x39 
#define __SWR 	0x3a 

#define __ADD 	0x40 
#define __ADDI 	0x41 
#define __ADDU 	0x42 
#define __ADDIU	0x43 
#define __SUB 	0x44 
#define __SUBU 	0x45 
#define __MULT 	0x46 
#define __MULTU	0x47 
#define __DIV 	0x48 
#define __DIVU 	0x49 
#define __MFHI 	0x4a 
#define __MTHI 	0x4b 
#define __MFLO 	0x4c 
#define __MTLO 	0x4d 
#define __AND 	0x4e 
#define __ANDI 	0x4f 

#define __OR 	0x50 
#define __ORI 	0x51 
#define __XOR 	0x52 
#define __XORI 	0x53 
#define __NOR 	0x54 
#define __SLL 	0x55 
#define __SLLV 	0x56 
#define __SRL 	0x57 
#define __SRLV 	0x58 
#define __SRA 	0x59 
#define __SRAV 	0x5a 
#define __SLT 	0x5b 
#define __SLTI 	0x5c 
#define __SLTU 	0x5d 
#define __SLTIU 0x5e 

#define __FADD_S 0x70 
#define __FADD_D 0x71 
#define __FSUB_S 0x72 
#define __FSUB_D 0x73 
#define __FMUL_S 0x74 
#define __FMUL_D 0x75 
#define __FDIV_S 0x76 
#define __FDIV_D 0x77 
#define __FABS_S 0x78 
#define __FABS_D 0x79 
#define __FMOV_S 0x7a 
#define __FMOV_D 0x7b 
#define __FNEG_S 0x7c 
#define __FNEG_D 0x7d

#define __CVT_S_D 0x80 
#define __CVT_S_W 0x81 
#define __CVT_D_S 0x82 
#define __CVT_D_W 0x83 
#define __CVT_W_S 0x84 
#define __CVT_W_D 0x85 

#define __C_EQ_S 0x90 
#define __C_EQ_D 0x91 
#define __C_LT_S 0x92 
#define __C_LT_D 0x93 
#define __C_LE_S 0x94 
#define __C_LE_D 0x95 
#define __FSQRT_S 0x96 
#define __FSQRT_D 0x97 

#define __NOP	0x00
#define __SYSCALL 0xa0
#define __BREAK 0xa1
#define __LUI 	0xa2
#define __MFC1 	0xa3 
#define __CFC1 	0xa4 
#define __MTC1 	0xa5 

#define __CTC1 	0xa6 
#define __DMFC1 0xa7 
#define __DMTC1 0xa8 

#define __LB_RR  0xc0 
#define __LBU_RR 0xc1 
#define __LHU_RR 0xc3 
#define __LW_RR  0xc4 
#define __L_S_RR 0xc5 
#define __SB_RR  0xc6 
#define __SH_RR  0xc7 
#define __SW_RR     0xc8 
#define __S_S_RR    0xc9 
#define __L_S_RR_R2 0xca 
#define __S_S_RR_R2 0xcb 
#define __LW_RR_R2  0xcc 
#define __SW_RR_R2  0xcd 
#define __DLW_RR    0xce 
#define __L_D_RR    0xcf 

#define __DSW_RR    0xd0 
#define __DSZ_RR    0xd1 
#define __S_D_RR    0xd2 
#define __LH_RR     0xc2 


#define ECOFF_EL_MAGIC          0x0162
#define ECOFF_STYP_TEXT		0x0020
#define ECOFF_STYP_RDATA	0x0100
#define ECOFF_STYP_DATA		0x0040
#define ECOFF_STYP_SDATA	0x0200
#define ECOFF_STYP_SBSS		0x0400
#define ECOFF_STYP_BSS		0x0080


#define	SYS_exit         1 
#define SYS_fork         2
#define	SYS_read	 3
#define	SYS_write	 4
#define	SYS_open	 5
#define	SYS_close	 6
#define	SYS_execv        11
#define	SYS_chdir        12
#define SYS_brk          17
#define	SYS_lseek	 19
#define	SYS_getpid       20
#define	SYS_ioctl	 54
#define	SYS_fstat	 62
#define	SYS_getpagesize	 64
#define SYS_sigvec       108
#define	SYS_gettimeofday 116
#define	SYS_getrusage	 117

/*****  3. macro definition *****/
#define Imm ((int)((short)(as->ir.low & 0xffff)))
#define UImm              (as->ir.low & 0xffff)
/* #define Shamt             (as->ir.low & 0xff) */
#define Shamt (as->ir.low & 0x1f) 
#define Rd_i (as->ir.low>>8  & 0xFF)
#define Rt_i (as->ir.low>>16 & 0xFF)
#define Rs_i (as->ir.low>>24 & 0xFF)
#define Rd (as->r[as->ir.low>>8  & 0xFF])
#define Fd (as->f[as->ir.low>>8  & 0xFF])
#define Rt (as->r[as->ir.low>>16 & 0xFF])
#define Rs (as->r[as->ir.low>>24 & 0xFF])
#define Cpc as->pc
#define Npc as->npc
#define Target (as->ir.low & 0x3ffffff)

#define TABLE_INDEX(addr) (((unsigned int)(addr)) >> 16 & 0x7fff)
#define SS_NFLAGS (sizeof(ss_flag_table)/sizeof(ss_flag_table[0]))

/***** 4. structure definitions *****/
struct evaluate{
  FILE *debug;
  ULL end_tc;
  unsigned int hi;
  unsigned int lo;
  ULL tc; /* retired instrunctions, total count */
  ULL icount[256];
};

struct architecture_state{ 
  ADDR_TYPE pc;     /* program counter      32-bit width. */
  ADDR_TYPE npc;    /* next program counter 32-bit width. */
  INST_TYPE ir;
  DATA_TYPE r[32];  /* integer register file         */
  DATA_TYPE f[32];  /* floating point register file  */
  DATA_TYPE hi;
  DATA_TYPE lo;
  DATA_TYPE fcc;
  char *block_table[BLOCK_TABLE_SIZE];

  ADDR_TYPE brk;              /* max address of the data segment */
  ADDR_TYPE text_min_address; /* min address of the text segment */
  ADDR_TYPE text_max_address; /* max address of the text segment */
};

struct machine_setting{
  int debug_flag;
  int trace_flag;
  int distrib_flag;
  int syscall_flag;
  int icount_flag;
  ULL end_tc;
  ULL trace_start;
  ULL trace_end;
};

struct ss_timeval{
  int ss_tv_sec;
  int ss_tv_usec;
};

struct ss_timezone{
  int ss_tz_minuteswest;
  int ss_tz_dsttime;
};

struct ss_rlimit{
  int ss_rlim_cur;
  int ss_rlim_max;
};

struct ss_flag_table{
  int ss_flag;
  int local_flag;
};

struct ss_rusage{
  struct ss_timeval ss_ru_utime;
  struct ss_timeval ss_ru_stime;
  int ss_ru_maxrss;
  int ss_ru_ixrss;
  int ss_ru_idrss;
  int ss_ru_isrss;
  int ss_ru_minflt;
  int ss_ru_majflt;
  int ss_ru_nswap;
  int ss_ru_inblock;
  int ss_ru_oublock;
  int ss_ru_msgsnd;
  int ss_ru_msgrcv;
  int ss_ru_nsignals;
  int ss_ru_nvcsw;
  int ss_ru_nivcsw;
};

struct  ss_statbuf{
  short          ss_st_dev;
  unsigned int   ss_st_ino;
  unsigned short ss_st_mode;
  short          ss_st_nlink;
  short          ss_st_uid;
  short          ss_st_gid;
  short          ss_st_rdev;
  int            ss_st_size;
  int            ss_st_atime;
  int            ss_st_spare1;
  int          	 ss_st_mtime;
  int            ss_st_spare2;
  int          	 ss_st_ctime;
  int            ss_st_spare3;
  int            ss_st_blksize;
  int            ss_st_blocks;
  unsigned int   ss_st_gennum;
  int            ss_st_spare4;
};

struct ecoff_filehdr{
  unsigned short f_magic;
  unsigned short f_nscns;
  int f_timdat;
  int f_symptr;
  int f_nsyms;
  unsigned short f_opthdr;
  unsigned short f_flags;
};

struct ecoff_aouthdr{
  short magic;
  short vstamp;
  int tsize;
  int dsize;
  int bsize;
  int entry;
  int text_start;
  int data_start;
  int bss_start;
  int gprmask;
  int cprmask[4];
  int gp_value;
};

struct ecoff_scnhdr{
  char s_name[8];
  int s_paddr;
  int s_vaddr;
  int s_size;
  int s_scnptr;
  int s_relptr;
  int s_lnnoptr;
  unsigned short s_nreloc;
  unsigned short s_nlnno;
  int s_flags;
};

/***** 5. function prototypes *****/
extern void fetch_instruction(ADDR_TYPE, INST_TYPE *);
extern void load_byte(int, ADDR_TYPE, DATA_TYPE*);
extern void load_2byte(int, ADDR_TYPE, DATA_TYPE*);
extern void load_4byte(ADDR_TYPE, DATA_TYPE*);
extern void load_nbyte(int, int, ADDR_TYPE, DATA_TYPE*);
extern void load_memory(ADDR_TYPE, char *, int);
extern void store_memory(ADDR_TYPE, char *, int);
extern void store_byte( ADDR_TYPE, DATA_TYPE);
extern void store_2byte(ADDR_TYPE, DATA_TYPE);
extern void store_4byte(ADDR_TYPE, DATA_TYPE);
extern void store_nbyte(int, ADDR_TYPE, DATA_TYPE);

extern void execute_instruction(struct architecture_state *);
extern void execute_syscall(struct architecture_state *);
extern void debugmode(struct architecture_state *);
extern void getasm(struct architecture_state *, char *);
extern void print_status(struct architecture_state *);
extern void mem_init(int, char **, char **);
extern void getopcode(int, char*);
extern void finish_simulator();
