/* execute.c                          written by Kenji Kise, 2001
 * A part of SimpleScalar Simulator.          Developing on Linux
 ****************************************************************/
#include <stdio.h>
#include <math.h>
#include "define.h"

extern struct machine_setting ms;
extern struct evaluate e;

void execute_instruction(struct architecture_state *as){
  float FVd, FVs, FVt;
  double FDd, FDs, FDt;
  memcpy(&FVs, &as->f[Rs_i], 4);
  memcpy(&FVt, &as->f[Rt_i], 4);
  memcpy(&FDs, &as->f[Rs_i], 8);
  memcpy(&FDt, &as->f[Rt_i], 8);

  switch(as->ir.op & 0xff){
  case __JALR: Rd=Cpc+8; Npc=Rs; break;
  case __JR:   Npc=Rs; break;
  case __J:    Npc=(Cpc&0xf0000000)|(Target<<2); break;
  case __JAL:  Npc=(Cpc&0xf0000000)|(Target<<2); as->r[31]=Cpc+8;
    break;

  case __BEQ:  if(Rs==Rt)   Npc=Cpc+8+(Imm<<2); break;
  case __BNE:  if(Rs!=Rt)   Npc=Cpc+8+(Imm<<2); break;
  case __BLEZ: if(Rs<=Rt)   Npc=Cpc+8+(Imm<<2); break;
  case __BGTZ: if(Rs> Rt)   Npc=Cpc+8+(Imm<<2); break;
  case __BLTZ: if(Rs< Rt)   Npc=Cpc+8+(Imm<<2); break;
  case __BGEZ: if(Rs>=Rt)   Npc=Cpc+8+(Imm<<2); break;
  case __BC1F: if(!as->fcc) Npc=Cpc+8+(Imm<<2); break;
  case __BC1T: if( as->fcc) Npc=Cpc+8+(Imm<<2); break;
    
  case __LB:  load_nbyte(1, SIGNED, Rs+Imm, &Rt); break;
  case __LH:  load_nbyte(2, SIGNED, Rs+Imm, &Rt); break;
  case __LW:  load_nbyte(4, SIGNED, Rs+Imm, &Rt); break;
  case __LBU: load_nbyte(1, UNSIGN, Rs+Imm, &Rt); break;
  case __LHU: load_nbyte(2, UNSIGN, Rs+Imm, &Rt); break;
  case __DLW: 
    load_nbyte(4, SIGNED, Rs+Imm,   &as->r[Rt_i]);
    load_nbyte(4, SIGNED, Rs+Imm+4, &as->r[Rt_i+1]);
    break;

  case __L_S: 
    load_nbyte(4, SIGNED, Rs+Imm, &as->f[Rt_i]); 
    break;
  case __L_D: 
    load_nbyte(4, SIGNED, Rs+Imm, &as->f[Rt_i]);
    load_nbyte(4, SIGNED, Rs+Imm+4, &as->f[Rt_i+1]);
    break;

  case __SB: store_nbyte(1, Imm+Rs, Rt); break;
  case __SH: store_nbyte(2, Imm+Rs, Rt); break;
  case __SW: store_nbyte(4, Rs+Imm, Rt); break;

  case __DSW:
    store_nbyte(4, Rs+Imm,   as->r[Rt_i]);
    store_nbyte(4, Rs+Imm+4, as->r[Rt_i+1]);
    break;

  case __S_S:
    store_nbyte(4, Rs+Imm, as->f[Rt_i]);
    break;

  case __S_D:
    store_nbyte(4, Imm+Rs,   as->f[Rt_i]);
    store_nbyte(4, Imm+Rs+4, as->f[Rt_i+1]);
    break;

  case __DSZ:
    store_nbyte(4, Rs+Imm,   0);
    store_nbyte(4, Rs+Imm+4, 0);
    break;

    /* You shold not use LWL/LWR/SWL/SWR, it is not safe. */
  case __LWL: load_nbyte(4, SIGNED, (Rs+Imm) & ~0x3, &Rt); break;
  case __LWR: load_nbyte(4, SIGNED, (Rs+Imm) & ~0x3, &Rt); break;
  case __SWL:
    {
      DATA_TYPE tmp, dat=0;
      load_nbyte(4, SIGNED, (Rs+Imm) & ~0x3, &tmp);
      switch(((Rs+Imm) & 0x03)){
      case 0: dat=(tmp & 0xFFFFFF00)|(Rt>>24 & 0x0000FF); break;
      case 1: dat=(tmp & 0xFFFF0000)|(Rt>>16 & 0x00FFFF); break;
      case 2: dat=(tmp & 0xFF000000)|(Rt>> 8 & 0xFFFFFF); break;
      case 3: dat=Rt; break;
      default:   printf("SWL address error\n");
      }
      store_nbyte(4, (Rs+Imm) & ~0x3, dat);
    }
    break;
  case __SWR:
    {
      DATA_TYPE tmp, dat=0;
      load_nbyte(4, SIGNED, (Rs+Imm) & ~0x3, &tmp);
      switch(((Rs+Imm) & 0x03)){
      case 0: dat=Rt; break;
      case 1: dat=(tmp & 0x000000FF)|(Rt<< 8 & 0xFFFFFF00); break;
      case 2: dat=(tmp & 0x0000FFFF)|(Rt<<16 & 0xFFFF0000); break;
      case 3: dat=(tmp & 0x00FFFFFF)|(Rt<<24 & 0xFF000000); break;
      default: printf("SWR address error\n");
      }
      store_nbyte(4, (Rs+Imm) & ~0x3, dat);
    }
    break;

  case __ADD:   Rd = Rs + Rt;                        break;
  case __ADDI:  Rt = Rs + Imm;                       break;
  case __ADDU:  Rd = Rs + Rt;                        break;
  case __ADDIU: Rt = Rs + Imm;                       break;
  case __SUB:   Rd = Rs - Rt;                        break;
  case __SUBU:  Rd = Rs - Rt;                        break;
  case __MFHI:  Rd = as->hi;                         break;
  case __AND:   Rd = Rs & Rt;                        break;
  case __ANDI:  Rt = Rs & UImm;                      break;
  case __OR:    Rd = Rs | Rt;                        break;
  case __ORI:   Rt = Rs | UImm;                      break;
  case __XOR:   Rd = Rs ^ Rt;                        break;
  case __XORI:  Rt = Rs ^ UImm;                      break;
  case __NOR:   Rd = ~(Rs | Rt);                     break;
  case __SLL:   Rd = Rt << Shamt;                    break;
  case __SLLV:  Rd = Rt << (Rs & 0x1f);              break;
  case __SRL:   Rd = (unsigned)Rt >> Shamt;          break;
  case __SRLV:  Rd = (unsigned)Rt >> (Rs & 0x1f);    break;
  case __SRA:   Rd = Rt >> Shamt;                    break;
  case __SRAV:  Rd = Rt >> (Rs & 0x1f);              break;
  case __SLT:   Rd = (Rs < Rt);                      break;
  case __SLTI:  Rt = (Rs < Imm);                     break;
  case __SLTU:  Rd = ((unsigned)Rs < (unsigned) Rt); break;
  case __SLTIU: Rt = ((unsigned)Rs < (unsigned)Imm); break;

  case __DIV:
    as->lo = Rs / Rt;
    as->hi = Rs % Rt;
    break;

  case __DIVU:
    as->lo = (unsigned)Rs / (unsigned)Rt;
    as->hi = (unsigned)Rs % (unsigned)Rt;
    break;

  case __MULT:
    {
      long long out;
      out =(long long)Rs * (long long)Rt;
      as->lo = (int)out;
      as->hi = (int)(out >> 32);
    }
    break;
    
  case __MULTU:
    {
      ULL out;
      out =((ULL)Rs & 0xFFFFFFFF) * ((ULL)Rt & 0xFFFFFFFF);
      as->lo = (unsigned int)out;
      as->hi = (unsigned int)(out >> 32);
    }
    break;
    
  case __MTHI: as->hi = Rs; break;
  case __MTLO: as->lo = Rs; break;
  case __MFLO: Rd = as->lo; break;

  case __FADD_S: FVd = FVs + FVt; memcpy(&Fd, &FVd, 4); break;
  case __FADD_D: FDd = FDs + FDt; memcpy(&Fd, &FDd, 8); break;
  case __FSUB_S: FVd = FVs - FVt; memcpy(&Fd, &FVd, 4); break;
  case __FSUB_D: FDd = FDs - FDt; memcpy(&Fd, &FDd, 8); break;
  case __FMUL_S: FVd = FVs * FVt; memcpy(&Fd, &FVd, 4); break;
  case __FMUL_D: FDd = FDs * FDt; memcpy(&Fd, &FDd, 8); break;
  case __FDIV_S: FVd = FVs / FVt; memcpy(&Fd, &FVd, 4); break;
  case __FDIV_D: FDd = FDs / FDt; memcpy(&Fd, &FDd, 8); break;
  case __FMOV_S: FVd = FVs;       memcpy(&Fd, &FVd, 4); break;
  case __FMOV_D: FDd = FDs;       memcpy(&Fd, &FDd, 8); break;
  case __FNEG_S: FVd =-FVs;       memcpy(&Fd, &FVd, 4); break;
  case __FNEG_D: FDd =-FDs;       memcpy(&Fd, &FDd, 8); break;
  case __FABS_S: FVd = fabs((double)FVs); memcpy(&Fd, &FVd, 4); break;
  case __FSQRT_S:FDd = sqrt((double)FDs); memcpy(&Fd, &FDd, 4); break;
  case __FABS_D: FDd = fabs(FDs); memcpy(&Fd, &FDd, 8); break;
  case __FSQRT_D:FDd = sqrt(FDs); memcpy(&Fd, &FDd, 8); break;
    
  case __CVT_S_D: FVd = (float )FDs; memcpy(&Fd, &FVd, 4); break;
  case __CVT_D_S: FDd = (double)FVs; memcpy(&Fd, &FDd, 8); break;
  case __CVT_S_W: FVd = (float )as->f[Rs_i]; memcpy(&Fd,&FVd,4);break;
  case __CVT_D_W: FDd = (double)as->f[Rs_i]; memcpy(&Fd,&FDd,8);break;
  case __CVT_W_S: as->f[Rd_i] = (int)FVs; break;    
  case __CVT_W_D: as->f[Rd_i] = (int)FDs; break;

  case __C_EQ_S: as->fcc=(FVs==FVt); break;
  case __C_EQ_D: as->fcc=(FDs==FDt); break;
  case __C_LT_S: as->fcc=(FVs <FVt); break;
  case __C_LT_D: as->fcc=(FDs <FDt); break;
  case __C_LE_S: as->fcc=(FVs<=FVt); break;
  case __C_LE_D: as->fcc=(FDs<=FDt); break;

  case __NOP:    break;
  case __SYSCALL: execute_syscall(as);    break;
  case __BREAK:   fprintf(stderr, "__BREAK\n"); exit(0); break;
  case __LUI:     Rt = UImm << 16; break;
  case __MFC1:    Rt = as->f[Rs_i]; break;
  case __CFC1:    break;
  case __MTC1:    as->f[Rs_i] = Rt; break;
  case __CTC1:    break;
  case __DMFC1:
    Rt            = as->f[Rs_i];
    as->r[Rt_i+1] = as->f[Rs_i+1];
    break;
  case __DMTC1:
    as->f[Rs_i]   = as->r[Rt_i];
    as->f[Rs_i+1] = as->r[Rt_i+1] ;
    break;
    
  case __LB_RR:
  case __LBU_RR:
  case __LHU_RR:
  case __LW_RR:
  case __L_S_RR:
  case __SB_RR:
  case __SH_RR:
  case __SW_RR:
  case __S_S_RR:
  case __L_S_RR_R2:
  case __S_S_RR_R2:
  case __LW_RR_R2:
  case __SW_RR_R2:
  case __DLW_RR:
  case __L_D_RR:
  case __DSW_RR:
  case __DSZ_RR:
  case __S_D_RR:
  case __LH_RR :
  default:
    fprintf(stderr, "TC:%qd  0x%x Not Implemented.\n", 
	   e.tc, as->ir.op & 0xff);
    exit(0);
  }
}
