/* memory.c                           written by Kenji Kise, 2001
 * A part of SimpleScalar Simulator.          Developing on Linux
 ****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "define.h"

extern struct architecture_state as;
extern struct evaluate e;

void mem_init(int argc, char **argv, char **argp){
  FILE *fp;
  FILE *fp_tmp;
  int i,j;
  int *buf;
  char *cbuf;
  ADDR_TYPE sp;
  ADDR_TYPE argv_addr, argp_addr;
  char *envp[] = {
    "TERM=vt100",
    NULL
  };

  struct ecoff_filehdr fhdr;
  struct ecoff_aouthdr ahdr;
  struct ecoff_scnhdr  shdr;

  char *program_name = argv[0];

  /* initialize a page table */
  for(i=0; i<BLOCK_TABLE_SIZE; i++) as.block_table[i]=NULL;

  /* read executable headers */
  fp = fopen(program_name, "r");
  fp_tmp = fopen(program_name, "r");
  if(fp==NULL || fp_tmp==NULL){
    fprintf(stderr, "Bad file name: %s\n", program_name);
    exit(1);
  }
  
  fread(&fhdr, sizeof(struct ecoff_filehdr), 1, fp);
  fread(&ahdr, sizeof(struct ecoff_aouthdr), 1, fp);
  if(fhdr.f_magic!=ECOFF_EL_MAGIC){
    printf("bad magic number in executable %s.\n", program_name);
    exit(1);
  }

  /* read executable sections */
  for(i=0; i<fhdr.f_nscns; i++){
    int text_size;

    fread(&shdr, sizeof(struct ecoff_scnhdr), 1, fp);

    switch(shdr.s_flags){
    case ECOFF_STYP_TEXT:
      text_size = ((shdr.s_vaddr + shdr.s_size) - SS_TEXT_BASE) 
	+ 128;
      buf = (unsigned int *)malloc(shdr.s_size);
      fseek(fp_tmp, shdr.s_scnptr, 0);
      fread(buf, shdr.s_size, 1, fp_tmp);
      for(j=0; j<(shdr.s_size/4); j=j+1){
	store_nbyte(4, shdr.s_vaddr+j*4,   buf[j  ]);
	store_nbyte(4, shdr.s_vaddr+j*4+4, buf[j+1]);
      }
      as.text_min_address=shdr.s_vaddr;
      as.text_max_address=shdr.s_vaddr+j*4;
      as.pc = ahdr.entry;
      free(buf);
      break;
      
    case ECOFF_STYP_RDATA: 
    case ECOFF_STYP_DATA:  
    case ECOFF_STYP_SDATA: 
      cbuf = (char *)calloc(shdr.s_size,1);
      fseek(fp_tmp, shdr.s_scnptr, 0);
      fread(cbuf, shdr.s_size, 1, fp_tmp);
      store_memory(shdr.s_vaddr, cbuf, shdr.s_size);
      free(cbuf);
      break;
      
    case ECOFF_STYP_SBSS:  break;
    case ECOFF_STYP_BSS:   break;
    }
  }    
  fclose(fp_tmp);
  fclose(fp);

  sp = STACK_BASE - MAX_ENVIRON;
  as.r[29] = sp;
  store_nbyte(4, sp, argc); sp +=4;

  argv_addr = sp;
  sp = sp + argc * sizeof(ADDR_TYPE);
  sp += sizeof(ADDR_TYPE); /* For NULL */

  argp_addr = sp;
  for(i=0; envp[i]; i++) sp += sizeof(ADDR_TYPE);
  sp += sizeof(ADDR_TYPE); /* For NULL */

  /** Store argv into simulator memory **/
  for(i=0; i<argc; i++){ 
    char c;
    DATA_TYPE d;
    char *ptr = argv[i];
    store_nbyte(4, argv_addr + i*sizeof(ADDR_TYPE), sp);
    do{
      c = *ptr++;
      d = (DATA_TYPE)c;
      store_nbyte(1, sp++, d);
    }while(c);
  }

  /** Store argp into simulator memory **/
  for(i=0; envp[i]; i++){ 
    char c;
    DATA_TYPE d;
    char *ptr = envp[i];
    store_nbyte(4, argp_addr + i*sizeof(ADDR_TYPE), sp);
    do{
      c = *ptr++;
      d = (DATA_TYPE)c;
      store_nbyte(1, sp++, d);
    }while(c);
  }


  /* Set Stack Pointer, sp (alias of $29) */
  as.brk  = 0x10004000;
}

char *allocblock(ADDR_TYPE adr){
  char *ret=as.block_table[TABLE_INDEX(adr)]=calloc(BLOCK_SIZE,1);
  if(ret==NULL){
    printf("Memory allocation error in allocblock.\n");
    exit(0);
  }
  return ret;
}

void store_memory(ADDR_TYPE adr, char *buf, int n){
  int i;
  for(i=0; i<n; i++){
    store_byte(adr+i, *(buf+i));
  }
}
void load_memory(ADDR_TYPE adr, char *buf, int n){
  int i;
  DATA_TYPE d;
  for(i=0; i<n; i++){
    load_byte(SIGNED, adr+i, &d);
    *(buf+i)= (char)(d & 0xFF);
  }
}
void fetch_instruction(ADDR_TYPE pc, INST_TYPE *ir){
  if(as.text_min_address>pc || as.text_max_address<pc){
    printf("PC:0x%08x wrong fetch address in load_inst\n",pc);
    exit(1);
  }
  load_nbyte(4, SIGNED, pc  , &ir->op);
  load_nbyte(4, SIGNED, pc+4, &ir->low);
}

void load_byte(int fnc, ADDR_TYPE adr, DATA_TYPE *data){
  char c;
  char *ptr = as.block_table[TABLE_INDEX(adr)];
  unsigned int offset = adr & BLOCK_MASK;
  if(ptr==NULL) ptr=allocblock(adr);
  c = *(ptr+offset);
  if(fnc==SIGNED) *data = (DATA_TYPE)c;
  else *data = (DATA_TYPE)c & 0xFF;
}

void load_2byte(int fnc, ADDR_TYPE adr, DATA_TYPE *data){
  short srt;
  char *ptr = as.block_table[TABLE_INDEX(adr)];
  unsigned int offset = adr & BLOCK_MASK;
  if(adr%2!=0) printf("load_2byte miss-alignment.\n");
  if(ptr==NULL) ptr=allocblock(adr);
  memcpy(&srt, ptr+offset, 2);
  if(fnc==SIGNED)  *data = srt;
  else *data = (int)srt & 0xFFFF;
}

void load_4byte(ADDR_TYPE adr, DATA_TYPE *data){
  char *ptr = as.block_table[TABLE_INDEX(adr)];
  unsigned int offset = adr & BLOCK_MASK;
  if(adr%4!=0) printf("%qd load_4byte miss-alignment.\n", e.tc);
  if(ptr==NULL) ptr=allocblock(adr);
  memcpy(data, ptr+offset, 4);
}

void store_byte(ADDR_TYPE adr, DATA_TYPE data){
  char c = (char)(data & 0xFF);
  char *ptr = as.block_table[TABLE_INDEX(adr)];
  unsigned int offset = adr & BLOCK_MASK;
  if(ptr==NULL) ptr=allocblock(adr);
  *(ptr+offset) = c;
}

void store_2byte(ADDR_TYPE adr, DATA_TYPE data){
  short srt = (short)(data & 0xFFFF);
  char *ptr = as.block_table[TABLE_INDEX(adr)];
  unsigned int offset = adr & BLOCK_MASK;
  if(ptr==NULL) ptr=allocblock(adr);
  memcpy(ptr+offset, &srt, 2);
}

void store_4byte(ADDR_TYPE adr, DATA_TYPE data){
  char *ptr = as.block_table[TABLE_INDEX(adr)];
  unsigned int offset = adr & BLOCK_MASK;
  if(adr%4!=0) printf("store_4byte miss-alignment.\n");
  if(ptr==NULL) ptr=allocblock(adr);
  memcpy(ptr+offset, &data, 4);
}

void store_nbyte(int n, ADDR_TYPE adr, DATA_TYPE data){
  switch(n){
  case 1: store_byte( adr, data); break;
  case 2: store_2byte(adr, data); break;
  case 4: store_4byte(adr, data); break;
  default:
    printf("Case %d, Error in store_nbyte\n", n); exit(1);
  }
}

void load_nbyte(int n, int fnc, ADDR_TYPE adr, DATA_TYPE *data){
  switch(n){
  case 1: load_byte( fnc, adr, data); break;
  case 2: load_2byte(fnc, adr, data); break;
  case 4: load_4byte(     adr, data); break;
  default:
    printf("Case %d, Error in load_nbyte\n", n); exit(1);
  }
}
