/* simu.c                             written by Kenji Kise, 2001
 * A part of SimpleScalar Simulator.          Developing on Linux
 ****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "define.h"

static char UsageMessage[] =
"SimpleScalar Simulator Ver 0.5a by Kenji KISE, Jan 2001\n\
Usage: SimuSimple [-option] executable [arguments]\n\
  -d: debug mode(set ms.debug_flag).\n\
  -s: print systemcall trace(set ms.syscall_flag).\n\
  -e[num][m]: finish after the num instructions run.\n\
  -i: create icount.dat\n\
  #-t: trace mode(set ms.trace_flag).\n\
";

struct architecture_state as;
struct machine_setting ms;
struct evaluate e;

void usage(){
  printf("%s", UsageMessage); 
  exit(0);
}

void finish_simulator(){
  int i;
  FILE *fp;
  char buf[1024];
  
  fprintf(stderr,"** %qd instructions are executed.\n", e.tc);
  fprintf(stderr,"** Simulation finished normally.\n");

  if(ms.icount_flag){
    fp = fopen("icount.dat", "w");
    for(i=0; i<256; i++){
      if(i==0x01) fprintf(fp, "** Control transfer instruction\n");
      if(i==0x20) fprintf(fp, "** Load instruction\n");
      if(i==0x30) fprintf(fp, "** Store instruction\n");
      if(i==0x40) fprintf(fp, "** Arithmetic instruction\n");
      if(i==0x50) fprintf(fp, "** Logical instruction\n");
      if(i==0x70) fprintf(fp, "** Floating-point instruction\n");
      if(i==0x80) fprintf(fp, "** Convert instruction\n");
      if(i==0x90) fprintf(fp, "** Fcc set instruction\n");
      if(i==0xa0) fprintf(fp, "** Systemcall Etc.\n");
      
      getopcode(i, buf);
      if(e.icount[i]) 
	fprintf(fp, "0x%2x %8s %10qd\n", i, buf, e.icount[i]);
    }
    fprintf(fp,"** %qd instructions are executed.\n", e.tc);
    fprintf(fp,"** Simulation finished normally.\n");
    fprintf(fp,"** icount.dat was created.\n");
    fclose(fp);
    fprintf(stderr,"** icount.dat was created.\n");
  }
  exit(0);
}

void init(int argc, char **argv, char **argp){
  int i;
  memset(&as, 0, sizeof(as));
  memset(&e , 0, sizeof(e));
  memset(&ms, 0, sizeof(ms));
  if(argc>1){
    for(i=0; i<10; i++){
      if(argc==1) break;
      if((*(char *)(argv[1]))=='-'){
	switch( *((char *)(argv[1]) +1) )
	  {
	  case 'h': usage();           break;
	  case 'd': ms.debug_flag=1;   break;
	  case 't': ms.trace_flag=1;   break;
	  case 's': ms.syscall_flag=1; break;
	  case 'i': ms.icount_flag=1;  break;
	  case 'e':
            ms.end_tc =atol((char *)(argv[1])+2);
            if(*((char *)(argv[1])+strlen(argv[1])-1)=='m')
              ms.end_tc *=1000000;
            printf("End tc: %.0f\n", (double)ms.end_tc);
            break;
	  default:
	    fprintf(stderr, "Illegal option -%c is used.\n",
		    *((char *)(argv[1]) +1));
	    exit(1);
	  }
	argv++;
	argc--;
      }
    }
  }
  if(argc < 2) usage();
  argv++;
  argc--;

  ms.trace_start = 0;
  ms.trace_end   = 20000;
  mem_init(argc, argv, argp);
}

void main_loop(){
  while(1){ 
    as.r[0]=0;
    fetch_instruction(as.pc, &as.ir); as.npc=as.pc+8;
    execute_instruction(&as);

    e.tc++;
    e.icount[as.ir.op & 0xff]++;

    if(ms.end_tc==e.tc) finish_simulator();
    if(ms.debug_flag) debugmode(&as); 
    if(ms.trace_flag && e.tc>ms.trace_start) print_status(&as);
    if(ms.trace_flag && e.tc==ms.trace_end) exit(0);

    as.pc = as.npc;
  }
}

int main(int argc, char **argv, char **argp){
  init(argc, argv, argp);
  main_loop();
  return 0;
}
