/* syscall.c                          written by Kenji Kise, 2001
 * A part of SimpleScalar Simulator.          Developing on Linux
 ****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/resource.h>
#include <unistd.h>
#include "define.h"

extern struct machine_setting ms;
extern struct evaluate e;

struct ss_flag_table ss_flag_table[] = {
  {0,00}, {1,01}, {2,02}, {0x0008,02000},
  {0x0200,0100}, {0x0400,01000}, {0x0800,0200},
  {0x4000,04000}, {0x8000,0400}, {0x2000,010000},
};

void sys_error_check(struct architecture_state *as){
  if(as->r[2]!=-1) as->r[7]=0;
  else{as->r[2]=errno; as->r[7]=1;}
}


void execute_syscall(struct architecture_state *as){
  int code = as->r[2];
  char *buf;

  switch(code){
  case	SYS_exit:
    finish_simulator();
    break;

  case SYS_getrusage:
    {
      struct rusage local_rusage;
      struct ss_rusage rusage;
      as->r[2]=getrusage((enum __rusage_who)as->r[4],
			 &local_rusage);
      sys_error_check(as);
      
      rusage.ss_ru_utime.ss_tv_sec = local_rusage.ru_utime.tv_sec;
      rusage.ss_ru_utime.ss_tv_usec= local_rusage.ru_utime.tv_usec;
      rusage.ss_ru_utime.ss_tv_sec = local_rusage.ru_utime.tv_sec;
      rusage.ss_ru_utime.ss_tv_usec= local_rusage.ru_utime.tv_usec;
      rusage.ss_ru_stime.ss_tv_sec = local_rusage.ru_stime.tv_sec;
      rusage.ss_ru_stime.ss_tv_usec= local_rusage.ru_stime.tv_usec;
      rusage.ss_ru_stime.ss_tv_sec = local_rusage.ru_stime.tv_sec;
      rusage.ss_ru_stime.ss_tv_usec= local_rusage.ru_stime.tv_usec;
      rusage.ss_ru_maxrss  = local_rusage.ru_maxrss;
      rusage.ss_ru_ixrss   = local_rusage.ru_ixrss;
      rusage.ss_ru_idrss   = local_rusage.ru_idrss;
      rusage.ss_ru_isrss   = local_rusage.ru_isrss;
      rusage.ss_ru_minflt  = local_rusage.ru_minflt;
      rusage.ss_ru_majflt  = local_rusage.ru_majflt;
      rusage.ss_ru_nswap   = local_rusage.ru_nswap;
      rusage.ss_ru_inblock = local_rusage.ru_inblock;
      rusage.ss_ru_oublock = local_rusage.ru_oublock;
      rusage.ss_ru_msgsnd  = local_rusage.ru_msgsnd;
      rusage.ss_ru_msgrcv  = local_rusage.ru_msgrcv;
      rusage.ss_ru_nsignals= local_rusage.ru_nsignals;
      rusage.ss_ru_nvcsw   = local_rusage.ru_nvcsw;
      rusage.ss_ru_nivcsw  = local_rusage.ru_nivcsw;
      
      store_memory(as->r[5], (char *)&rusage, 
		   sizeof(struct ss_rusage));
    }
    break;
    
  case SYS_gettimeofday:
    {
      struct ss_timeval ss_tv;
      struct ss_timezone ss_tz;
      struct timezone tz, *tzp;
      struct timeval tv, *tvp;
      
      if(as->r[4]!=0){
	load_memory(as->r[4], (char *)&ss_tv, 
		    sizeof(struct ss_timeval));
	tv.tv_sec  = ss_tv.ss_tv_sec;
	tv.tv_usec = ss_tv.ss_tv_usec;
	tvp = &tv;
      }
      else tvp=0;
      
      if (as->r[5]!=0){
	load_memory(as->r[5], (char *)&ss_tz, 
		    sizeof(struct ss_timezone));
	tz.tz_minuteswest = ss_tz.ss_tz_minuteswest;
	tz.tz_dsttime     = ss_tz.ss_tz_dsttime;
	tzp = &tz;
      }
      else tzp=0;
      
      as->r[2]=gettimeofday(tvp, tzp);
      sys_error_check(as);
      
      if(as->r[4]!=0){
	ss_tv.ss_tv_sec  = tv.tv_sec;
	ss_tv.ss_tv_usec = tv.tv_usec;
	store_memory(as->r[4], (char *)&ss_tv, 
		     sizeof(struct ss_timeval));
      }
      if(as->r[5] != 0){
	ss_tz.ss_tz_minuteswest = tz.tz_minuteswest;
	ss_tz.ss_tz_dsttime     = tz.tz_dsttime;
	store_memory(as->r[5], (char *)&ss_tz, 
		     sizeof(struct ss_timezone));
      }
    }
    break;

  case SYS_sigvec:
    as->r[2]=as->r[7]=0;
    break;

  case SYS_getpagesize:
    as->r[2] = PAGE_SIZE;
    break;

  case SYS_read:
    buf = (char *)calloc(as->r[6], 1);
    as->r[2] = read(as->r[4], buf, as->r[6]);
    sys_error_check(as);
    store_memory(as->r[5], buf, as->r[2]);
    free(buf);
    break;
    
  case SYS_write:
    buf=(char *)calloc(as->r[6], 1);
    load_memory(as->r[5], buf, as->r[6]);
    as->r[2] = write(as->r[4], buf, as->r[6]);
    fflush(stdout);
    if(as->r[2]==as->r[6]) as->r[7]=0;
    else{as->r[2]=errno; as->r[7]=1;}
    free(buf);
    break;

  case SYS_open:
    {
      char buf[1024];
      unsigned int i;
      DATA_TYPE d;
      int ss_flags = as->r[5];
      int local_flags = 0;
     
      for(i=0; i<SS_NFLAGS; i++){
	if(ss_flags & ss_flag_table[i].ss_flag){
	  ss_flags    &= ~ss_flag_table[i].ss_flag;
	  local_flags |=  ss_flag_table[i].local_flag;
	}
      }

      for(i=0; i<1024; i++){
	load_nbyte(1, SIGNED, as->r[4]+i, &d);
	*(buf+i) = (char)(d & 0xFF);
	if(*(buf+i)==0) break;
      }	
      as->r[2]=open(buf, local_flags, as->r[6]);
      sys_error_check(as);
    }
    break;

  case SYS_close:
    if(as->r[4]==0 || as->r[4]==1 || as->r[4]==2){
      as->r[7]=0;
    }
    else{
      as->r[2]=close(as->r[4]);
      sys_error_check(as);
    }
    break;

  case SYS_lseek:
    as->r[2]=lseek(as->r[4], as->r[5], as->r[6]);
    sys_error_check(as);
    break;
    
  case SYS_getpid:
    as->r[2]=getpid();
    sys_error_check(as);
    break;
    
  case SYS_brk:
    {
      ADDR_TYPE addr=(as->r[4]+PAGE_SIZE-1) & ~(PAGE_SIZE-1);
      if(addr>=as->brk && addr<(unsigned int)as->r[29]){
	as->r[2]=0;
	as->r[7]=0;
	as->brk=addr;
      }
      else{
	as->r[2]=ENOMEM;
	as->r[7]=1;
      }
    }
    break;
    
  case SYS_ioctl:
    {
      char buf[128];
      int local_req=0;
      
      switch(as->r[5]){
      case  1074164744: local_req = 0x5481; break;
      case -2147060727:	local_req = 0x5482; break;
      case  1076130901: local_req = 0x5481; break;
      case  1075082331: local_req = 0x5405; break;
      case  1074164852: local_req = 0x5485; break;
      case -2147060619: local_req = 0x5486; break;
      case  1074295912: local_req = 0x5413; break;
      case -2146143143: local_req = 0x5407; break;
      case  1074164754: local_req = 0x5483; break;
      case -2147060719: local_req = 0x5484; break;
      case  0x4004747c: local_req = 0x5487; break;
      case  0x8004747d: local_req = 0x5488; break;
      }
      
      if(!local_req){
	printf("syscall: ioctl error.\n"); exit(0);
      }
      else{
	if(as->r[6]!=0) load_memory(as->r[6], buf, 128); 
	as->r[2]=ioctl(as->r[4], local_req, buf);
	if(as->r[6]!=0) store_memory(as->r[6], buf, 128);
	sys_error_check(as);
      }
    }
    break;
    
  case SYS_fstat:
    {
      struct ss_statbuf ss_sbuf;
      struct stat sbuf;
      
      as->r[2]=fstat(as->r[4], &sbuf);
      sys_error_check(as);
      
      ss_sbuf.ss_st_dev     = sbuf.st_dev;
      ss_sbuf.ss_st_ino     = sbuf.st_ino;
      ss_sbuf.ss_st_mode    = sbuf.st_mode;
      ss_sbuf.ss_st_nlink   = sbuf.st_nlink;
      ss_sbuf.ss_st_uid     = sbuf.st_uid;
      ss_sbuf.ss_st_gid     = sbuf.st_gid;
      ss_sbuf.ss_st_rdev    = sbuf.st_rdev;
      ss_sbuf.ss_st_size    = sbuf.st_size;
      ss_sbuf.ss_st_atime   = sbuf.st_atime;
      ss_sbuf.ss_st_mtime   = sbuf.st_mtime;
      ss_sbuf.ss_st_ctime   = sbuf.st_ctime;
      ss_sbuf.ss_st_blksize = sbuf.st_blksize;
      ss_sbuf.ss_st_blocks  = sbuf.st_blocks;
      store_memory(as->r[5], (char*)&ss_sbuf, 
		   sizeof(struct ss_statbuf));
    }
    break; 

  default:
    printf("Tc:%8qd  In syscall.c Default Error %d\n", 
	   e.tc, as->r[2]);
    exit(1);
  }

  if(ms.syscall_flag){
    printf("** TC:%qd: %d syscall(%d,%d,%d) => %d \n", 
	   e.tc, code, as->r[4], as->r[5], as->r[6], as->r[2]);
    fflush(stdout);
  }
}
